"""
Nanugo allows you to generate Anki decks from PDF files using both an easy-to-use CLI or as a Python module. You can think of it as creating a real-life flashcard—similar to folding (or, in this case, cutting) a piece of paper in half (or with recent update, in multiple pieces!)

## Quickstart
```shell
$ nanugo /path/to/pdf

# Check out all the functionalities with
$ nanugo --help
```

"""

__docformat__ = "google"

from .tools import images
from . import nanugo, log



from typing import List, Dict, Tuple
import pypdfium2 as pdfium
import PIL, logging, warnings, os
from ..log import logger

__docformat__ = "google"


def convert_pdf(
    pdf_path: str, render_scale: float = 1
) -> List[Dict[int, PIL.Image.Image]]:
    """Converts PDF files to PIL.Image.Image object.

    Args:
        pdf_path (str): Path for the pdf file to be converted
        render_scale (int): Render scale to use with pdfium.PdfDocument.render() function.

    Returns:
        List[Dict[int, PIL.Image.Image]]: List of PIL.Image.Image objects.
    """
    result = []

    pdf_file = pdfium.PdfDocument(pdf_path)
    page_indicies = [i for i in range(len(pdf_file))]

    # If your code shall be frozen into an executable, multiprocessing.freeze_support() needs to be called at the start of the if __name__ == "__main__": block if using this method.
    renderer = pdf_file.render(
        pdfium.PdfBitmap.to_pil, page_indices=page_indicies, scale=render_scale
    )

    for image in renderer:
        result.append(image)

    logger.debug(
        f"{len(result)} pages has been successfully rendered: {os.path.split(pdf_path)[-1]}"
    )

    return result


def split_page(
    image: PIL.Image.Image,
    vertical: bool = False,
    ratio: tuple = (0.5, 0.5),
    rows: int = 1,
) -> List[Tuple[PIL.Image.Image, PIL.Image.Image]]:
    """Splits given PIL.Image.Image object into two.

    Args:
        image (PIL.Image.Image): PIL.Image.Image object to be split
        vertical (bool, optional): if true, the image will be split vertically, not horizontally. Defaults to False.
        ratio (tuple, optional): ratio to which the image will be split. Defaults to (0.5, 0.5).
        rows (int, optional): if given value greater than 1, pages will be split by multiple times horizontally and then each vertically.

    Returns:
        List[Tuple[PIL.Image.Image, PIL.Image.Image]]: A list of tuple(s) of two PIL.Image.Image objects. First Image will going to be the top/left one by default.
    """
    d_width, d_height = image.size  # Get default sizes

    if sum(ratio) != 1:
        warnings.warn(
            "Custom ratio was given but the sum of it is not 1. If this is intended, you can safely ignore this warning.",
            UserWarning,
        )

    if rows == 1:
        if vertical:
            ratio_l, ratio_r = ratio
            image_l = image.crop((0, 0, d_width * ratio_l, d_height))
            image_r = image.crop((d_width * (1 - ratio_r), 0, d_width, d_height))

            return [(image_l, image_r)]

        else:
            ratio_t, ratio_b = ratio
            image_t = image.crop((0, 0, d_width, d_height * ratio_t))
            image_b = image.crop((0, d_height * (1 - ratio_b), d_width, d_height))

            return [(image_t, image_b)]

    elif rows > 1:
        ratio_l, ratio_r = ratio
        row_height = d_height // rows

        image_sets = []

        for i in range(rows):
            row_start = i * row_height
            row_end = (i + 1) * row_height if i < rows - 1 else d_height

            image_row = image.crop((0, row_start, d_width, row_end))

            image_row_l = image_row.crop((0, 0, d_width * ratio_l, row_height))
            image_row_r = image_row.crop(
                (d_width * (1 - ratio_r), 0, d_width, row_height)
            )

            image_sets.append((image_row_l, image_row_r))

        return image_sets



import os, sys
from glob import glob
from ..log import logger

__docformat__ = "google"


def dir_path(path: str) -> None:
    """Check if given path actually exists.

    Args:
        path (str): path for the validation

    Raises:
        NotADirectoryError
    """
    if os.path.isdir(path):
        return path
    else:
        raise NotADirectoryError(path)


def file_exists_prompt(export_file_path: str, pkg_name: str):
    """Check if file exists. If exists, ask for user input.

    Args:
        export_file_path (str): Path to be checked
        pkg_name (str): Name for the package.
    """
    if os.path.isfile(export_file_path):
        logger.debug(
            f"os.path.isfile(export_file_path) = {os.path.isfile(export_file_path)}"
        )
        answer = input(
            f"File {pkg_name}.apkg already exists! Do you want to proceed anyway? (y/n): "
        )
        if answer.lower() in ["y", "Y"]:
            pass
        else:
            logger.info("Anki package was not created.")
            sys.exit()


def pdf_file_arg(arg):
    """Check if the given path argument(and plain path) is valid and according file(s) exist(s).

    Args:
        arg : Path argument (can also contain wildcard)

    Raises:
        FileNotFoundError
    """
    if len(glob(arg)) == 0:
        raise FileNotFoundError(
            f"Cannot find any files that matches the following argument: {arg}"
        )
    else:
        pass



"""
CLI for nanugo. It's better to use `nanugo --help` than reading this.
"""
__docformat__ = "google"
from glob import glob
from .tools import validations
from .log import logger
from .nanugo import Builder
import pkg_resources
import argparse, sys, os, logging


def main(args=None):
    parsed_args = get_args(args)

    if parsed_args.debug:
        logger.setLevel(logging.DEBUG)

    logger.debug(parsed_args)

    pkg_name = parsed_args.package_name[0]
    export_file_path = os.path.join(parsed_args.dest, f"{pkg_name}.apkg")

    validations.file_exists_prompt(export_file_path, pkg_name)

    file_paths = []

    for arg in parsed_args.pdf_files:
        try:
            validations.pdf_file_arg(arg)
            file_paths += glob(arg)
        except:
            sys.exit()

    logger.info(f"Total {len(file_paths)} file(s) will be converted.")

    builder = Builder()
    decks = []

    file_name_only = lambda path: os.path.splitext(os.path.basename(path))[0]

    image_width = parsed_args.width if parsed_args.width else 0

    for file in file_paths:
        deck = builder.build_deck(
            file_name_only(file),
            file,
            ratio=tuple(parsed_args.ratio),
            vertical=parsed_args.vertical,
            inversed=parsed_args.inversed,
            rows=parsed_args.rows,
            image_width=image_width,
            render_scale=parsed_args.scale if parsed_args.scale else 1.0,
        )
        decks.append(deck)

    pkg = builder.build_pkg(decks)
    pkg.write_to_file(export_file_path)

    logger.info(f"Conversion completed: {export_file_path}")

    # No need for explicit return
    # return pkg


def get_args(args):
    parser = get_parser()
    return parser.parse_args(args)


def get_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        allow_abbrev=True,
        description="Turn your handwritten pdf sheets to Anki deck(s).",
        epilog="Created by @lewisleedev. If there's any problem, let me know at https://github.com/lewisleedev/nanugo/issues",
        formatter_class=argparse.RawTextHelpFormatter,
    )
    parser.add_argument(
        "package_name",
        action="store",
        type=str,
        nargs=1,
        help="File name for your .apkg file. Do not put file extension in the end.",
    )
    parser.add_argument(
        "pdf_files",
        help="Input PDF file to convert",
        nargs="*",
    )
    parser.add_argument(
        "-d,",
        "--dest",
        dest="dest",
        metavar="",
        action="store",
        type=validations.dir_path,
        default="./",
        help="Destination folder for your .apkg output. Defaults to current directory.",
    )
    parser.add_argument(
        "--ratio",
        dest="ratio",
        metavar="",
        action="store",
        type=float,
        nargs=2,
        default=(0.5, 0.5),
        help="Custom ratio to which each pages will be split. Defaults to (0.5 0.5). It's best to keep the sum of the ratios to 1. First number is the ratio for the top or left.",
    )
    parser.add_argument(
        "--inversed",
        "-i",
        dest="inversed",
        action="store_true",
        help="Inverses the Question/Answer side.",
    )
    parser.add_argument(
        "--vertical",
        "-v",
        dest="vertical",
        action="store_true",
        help="Splits vertically instead of horizontally",
    )
    parser.add_argument(
        "--debug",
        "--d",
        dest="debug",
        action="store_true",
        help="For even bigger nerds",
    )
    parser.add_argument(
        "--version",
        "-V",
        action="version",
        version=pkg_resources.get_distribution("nanugo").version,
        help="Shows version and exits.",
    )

    parser.add_argument(
        "--rows",
        type=int,
        default=1,
        help="Number of rows for vertical splitting (default: 1). By having rows value bigger than 1, each row will only be cut vertically.",
    )

    parser.add_argument(
        "--width",
        type=int,
        help="When given value greater than 1, images in question/answer side will be created with given value of css width property. Height will be set to auto. If you are trying to fix cards being too small, use higher scales value not width.",
    )

    parser.add_argument(
        "--scale",
        type=float,
        help="Scale for Pdfium.PdfPage.render() to use during pdf->image rendering process. Defaults to 1. Note than bigger numbers may cause (significant) workload. If you're having issues with blurry images of cards, higher scale value might help (e.g. 1.5).",
    )

    return parser


if __name__ == "__main__":
    main()



"""
Main module for `nanugo`. Using `Builder.build_deck` and `Builder.build_pkg` will suffice in most cases. Use with `genanki` package.
"""

__docformat__ = "google"

import genanki, tempfile, os, logging, itertools, shutil, sys
from .tools import images
from collections import namedtuple
from .log import logger
from .tools import utils

log = logging.getLogger(__name__)

BuilderDeck = namedtuple("BuilderDeck", "deck media_list")
BuilderDeck.__doc__ = """Builds Anki deck object with media file list.

Args:
    deck (genanki.Deck):built deck object
    media_list (list):paths to media files
"""


class Builder:
    """Builder for nanugo. __init__ Prepares Model and tempdir.."""

    def __init__(self):
        self.model = genanki.Model(
            # Number below is intended to be hardcoded. It is not nessesary to be changed. Leave it as-is unless needed.
            2015122512,
            "Nanugo Model",
            fields=[{"name": "QuestionSide"}, {"name": "AnswerSide"}],
            templates=[
                {
                    "name": "nanugo template",
                    "qfmt": "{{QuestionSide}}<br>",
                    "afmt": '{{FrontSide}}<hr id="answer">{{AnswerSide}}',
                }
            ],
        )
        self.tempdir = tempfile.mkdtemp(
            prefix="nanugo_"
        )  # This should be deleted manually...
        logger.debug(f"tempdir created: {self.tempdir}")

    def build_deck(
        self,
        deck_name: str,
        pdf_path: str,
        vertical: bool = False,
        ratio: tuple = (0.5, 0.5),
        inversed: bool = False,
        rows: int = 1,
        image_width: int = 0,
        render_scale: float = 1,
    ) -> BuilderDeck:
        """Builds Anki deck object with media file list.

        Args:
            deck_name (str): name of the deck
            pdf_path (str): path of the pdf file to make deck from.
            vertical (bool, optional): when enabled, pages will be split vertically and not horizontally. Defaults to False.
            ratio (tuple, optional): ratio to which each page will be split. If the sum of the ratios are not 1, it will raise a warning but not an error. If you intend to use non-sum-one numbers, you can ignore said warning. Defaults to (0.5, 0.5).
            inversed (bool, optional): when enabled, question side and answer side will be reversed, making top/left side the answer side and vice versa. Defaults to False.
            rows (int, optional): When given value greater than 1, multiple cards will be created from a single page of the given pdf file. Note that only vertical cuts are supported.
            image_width (int, optional): When given value greater than 1, images in question/answer side will be created with given value of css width property. Height will be set to auto. If you are trying to fix cards being too small, use higher scales value not width.
            render_scale (float, optional): Scale for Pdfium.PdfPage.render() to use during pdf->image rendering process. Defaults to 1. Note than bigger numbers may cause (significant) workload. Recommended to use only when you're using multiple rows.
        Returns:
            A `nanugo.BuilderDeck` object
        """
        if rows < 1:
            logger.error("Number of row(s) cannot be smaller than 1!")
            sys.exit()

        if rows > 1 and vertical == True:
            logger.warning(
                "Multiple row conversion selected. Vertical setting will be ignored."
            )

        logger.debug(image_width)

        # DeckID is also intended to be hardcoded.
        deck = genanki.Deck(1564947522, deck_name)
        logger.debug(f"Conversion started: {os.path.split(pdf_path)[-1]}")
        converted_pdf = images.convert_pdf(pdf_path, render_scale=render_scale)
        logger.debug(f"Splitting started: {os.path.split(pdf_path)[-1]}")
        split_pdf = list(
            map(
                lambda page: images.split_page(
                    page, vertical=vertical, ratio=ratio, rows=rows
                ),
                converted_pdf,
            )
        )
        logger.debug(f"Splitting finished: {os.path.split(pdf_path)[-1]}")
        media_list = []

        for set in split_pdf:
            for index, page in enumerate(
                set
            ):  # This is where you should probably add progress bar to.
                q_side, a_side = page[::-1] if inversed else page

                card_uid = utils.get_hash()

                q_side_path = os.path.join(
                    self.tempdir, f"{deck_name}_{card_uid}_q.jpeg"
                )
                a_side_path = os.path.join(
                    self.tempdir, f"{deck_name}_{card_uid}_a.jpeg"
                )
                q_side.save(q_side_path)
                logger.debug(f"Saved: {q_side_path}")
                a_side.save(a_side_path)
                logger.debug(f"Saved: {a_side_path}")

                media_list.extend([q_side_path, a_side_path])

                w_parsed = str(image_width) + "px" if image_width > 0 else "auto"

                note = genanki.Note(
                    model=self.model,
                    fields=[
                        f'<div style="text-align: center;"><img style="width: {w_parsed}; height: auto; margin: auto;" src="{deck_name}_{card_uid}_q.jpeg" /></div>',
                        f'<div style="text-align: center;"><img style="width: {w_parsed}; height: auto; margin: auto;" src="{deck_name}_{card_uid}_a.jpeg" /></div>',
                    ],
                )

                deck.add_note(note)
            logger.debug("Deck successfully created.")

        return BuilderDeck(deck, media_list)

    def build_pkg(self, builder_decks: list) -> genanki.Package:
        """Builds genaki.Package object with provided deck(s) and media file(s).

        Args:
            builder_decks (list): List of BuilderDeck object.

        Returns:
            genanki.Package: A genanki.Package object containing all the decks and list of media files.
        """
        decks, media_list = map(list, zip(*builder_decks))
        logger.info(f"Total {len(decks)} deck(s) will be added to the package.")
        pkg = genanki.Package(decks)
        pkg.media_files = itertools.chain(*media_list)
        return pkg

    def __del__(self):
        if os.path.exists(self.tempdir):
            shutil.rmtree(self.tempdir)
            logger.debug(
                f"tempdir {self.tempdir} was removed and the Builder object will be deleted."
            )
